

////////////////////////////////////////////////////////////////////////
//
// collision system interface
//
// NOTE: if using this coll system obj, it must be created
//       before you create your game objs. this is because
//       the game objs need to have a coll system active      
//       before they can start building their rigid bodies
//
////////////////////////////////////////////////////////////////////////


package si.game
{

	include "../siImportFile.as"


	public class CollSystem
	{
	

		private var world:b2World;

		private var ITERATIONS:int = 10;
		private var TIMESTEP:Number = 1/60;		
		private var PHYSICAL_SCALE:Number = 1;
		
		private var mouse_joint:b2MouseJoint;
		private var coll_mouse_x:Number;
		private var coll_mouse_y:Number;

		private var sprite_collmesh:Sprite;
	
		public var game:Game;	

		private var complex_create_bodydef:b2BodyDef;
		private var complex_create_boxdef:b2BoxDef;
		private var complex_create_circledef:b2CircleDef;
		private var complex_collision_category:uint;
		private var complex_collision_mask:uint;	
	

	
		////////////////////////////////////////////////////////////////////////
		//
		// 
		//
		////////////////////////////////////////////////////////////////////////
		public function CollSystem(game:Game, vec_grav:Point)
		{
			this.game = game;

			//initialize mouse stuff once before game start
			updateMouse();

			//sprite used for debug display of collision system meshes
			if (Game.DEBUG_COLL_SYSTEM)
			{
				sprite_collmesh = new Sprite();
				game.getStage().addChild(sprite_collmesh);
			}

			//build coll world
			var world_aabb:b2AABB = new b2AABB();
			world_aabb.minVertex.Set(0.0, 0.0);
			world_aabb.maxVertex.Set(600.0, 600.0);

			//set world gravity
			var grav:b2Vec2 = new b2Vec2(vec_grav.x, vec_grav.y);

			//system optimization - let sleeping bodies lie (i.e. not have calcs done on them)
			var do_sleep:Boolean = true;
	
			//build the world
			world = new b2World(world_aabb, grav, do_sleep);

		}
	

		////////////////////////////////////////////////////////////////////////
		//
		// 
		//
		////////////////////////////////////////////////////////////////////////
		public function shutdown():void
		{
			if (Game.DEBUG_COLL_SYSTEM)
			{
				game.getStage().removeChild(sprite_collmesh);
				sprite_collmesh = null;
			}

			world = null;			
		}




		///////////////////////////////////////////////////////////////////////////
		//
		//
		///////////////////////////////////////////////////////////////////////////
		public function update():void
		{
			updateMouse();

			if (mouse_joint)
			{
				var p2:b2Vec2 = new b2Vec2(coll_mouse_x, coll_mouse_y);
				mouse_joint.SetTarget(p2);
			}
			
			
			//update coll system
			world.Step(TIMESTEP, ITERATIONS);


			//update positions of sprites (as they are attached to the coll bodies
			for (var bb:b2Body = world.m_bodyList; bb; bb = bb.m_next)
			{
				if (bb.m_userData is Sprite)
				{
					bb.m_userData.x = bb.m_position.x;
					bb.m_userData.y = bb.m_position.y;
					bb.m_userData.rotation = bb.m_rotation * (180/Math.PI);
				}
			}			



			//draw coll joints and shapes
			if (Game.DEBUG_COLL_SYSTEM)
			{

				sprite_collmesh.graphics.clear()

				// joints
				for (var jj:b2Joint = world.m_jointList; jj; jj = jj.m_next)
					DrawJoint(jj);
				
				// bodies
				for (bb = world.m_bodyList; bb; bb = bb.m_next)
				{
					for (var s:b2Shape = bb.GetShapeList(); s != null; s = s.GetNext())
						DrawShape(s);
				}
			}
		}




		///////////////////////////////////////////////////////////////////////////
		//
		//
		///////////////////////////////////////////////////////////////////////////
		private function updateMouse():void
		{

			coll_mouse_x = (game.getStage().mouseX)/PHYSICAL_SCALE; 
			coll_mouse_y = (game.getStage().mouseY)/PHYSICAL_SCALE; 
			
		}


		///////////////////////////////////////////////////////////////////////////
		//
		//
		///////////////////////////////////////////////////////////////////////////
		public function createCollisionMesh(is_solid:Boolean, collision_category:uint, collision_mask:uint, size:Point, x:int, y:int, gobj:GameObj):b2Body
		{

			//build subpart to add to collision object
			//note: in this case we have only one - but could have many of these
			var boxDef:b2BoxDef;
			boxDef = new b2BoxDef();


			boxDef.extents.Set(size.x/2,size.y/2);
			boxDef.friction = 0;

			if (is_solid)
			{
				//static
				boxDef.density = 0;	
			}
			else
			{
				//dynamic
				boxDef.density = 1;	
			}

			//restitution contols the bounce of the object (0 means no bounce)
			boxDef.restitution = 1.0;

			boxDef.categoryBits = collision_category;
			boxDef.maskBits = collision_mask;


			//build actual top level coll obj
			var bodyDef:b2BodyDef;
			bodyDef = new b2BodyDef();
					
			//take parents position into account
			if (gobj.getParentObj() != game)
				bodyDef.position.Set(x + gobj.getParentObj().x,y + gobj.getParentObj().y);
			else		
				bodyDef.position.Set(x,y);

			bodyDef.userData = gobj;
			bodyDef.userData.width = boxDef.extents.x * 2;
			bodyDef.userData.height = boxDef.extents.y * 2;

			bodyDef.AddShape(boxDef);

			return world.CreateBody(bodyDef);
					
		}




		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function complexCreateStart(pos:Point,collision_category:uint, collision_mask:uint):void
		{
			complex_create_bodydef = new b2BodyDef();
			complex_create_bodydef.position.Set(pos.x,pos.y);
	
			complex_collision_category = collision_category;
			complex_collision_mask = collision_mask;		
		}
		

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function complexCreateAddBox(local_pos:Point, in_size:Point):void
		{
			complex_create_boxdef = new b2BoxDef();
			complex_create_boxdef.extents.Set(in_size.x,in_size.y);
			complex_create_boxdef.friction = 0;
			complex_create_boxdef.density = 0;	
			complex_create_boxdef.restitution = 0.0
			complex_create_boxdef.categoryBits = complex_collision_category;
			complex_create_boxdef.maskBits = complex_collision_mask;
			complex_create_boxdef.localPosition = new b2Vec2(local_pos.x,local_pos.y);

			complex_create_bodydef.AddShape(complex_create_boxdef);

		}

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function complexCreateAddCircle(local_pos:Point, in_size:Point):void
		{
			complex_create_circledef = new b2CircleDef();
			complex_create_circledef.radius = in_size.x;			
			complex_create_circledef.friction = 0;
			complex_create_circledef.density = 0;	
			complex_create_circledef.restitution = 0.0
			complex_create_circledef.categoryBits = complex_collision_category;
			complex_create_circledef.maskBits = complex_collision_mask;
			complex_create_circledef.localPosition = new b2Vec2(local_pos.x,local_pos.y);

			complex_create_bodydef.AddShape(complex_create_circledef);
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function complexCreateEnd():b2Body
		{
			return world.CreateBody(complex_create_bodydef);
		}





		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function destroyCollisionMesh(coll_body:b2Body):void
		{
			world.DestroyBody(coll_body);
		}
		

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function createMouseJoint(body:b2Body):void
		{
			var md:b2MouseJointDef = new b2MouseJointDef();
			md.body1 = world.m_groundBody;
			md.body2 = body;
			md.target.Set(body.m_position.x,body.m_position.y);

			md.maxForce = 30000.0 * body.m_mass;
			md.maxForce = 50000.0 * body.m_mass;
			md.timeStep = TIMESTEP;

			mouse_joint = world.CreateJoint(md) as b2MouseJoint;
			body.WakeUp();
		}

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function destroyMouseJoint():void
		{
			mouse_joint = null;
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function getContactList():b2Contact
		{
			return world.GetContactList();
		}



		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		private function DrawShape(shape:b2Shape):void
		{
			switch (shape.m_type)
			{
				case b2Shape.e_circleShape:
				{
					var circle:b2CircleShape = shape as b2CircleShape;
					var pos:b2Vec2 = circle.m_position;
					var r:Number = circle.m_radius;
					var k_segments:Number = 16.0;
					var k_increment:Number = 2.0 * Math.PI / k_segments;
					sprite_collmesh.graphics.lineStyle(1,0x000000,1);
					sprite_collmesh.graphics.moveTo((pos.x + r) * PHYSICAL_SCALE, (pos.y) * PHYSICAL_SCALE);
					var theta:Number = 0.0;
					
					for (var i:int = 0; i < k_segments; ++i)
					{
						var d:b2Vec2 = new b2Vec2(r * Math.cos(theta), r * Math.sin(theta));
						var v:b2Vec2 = b2Math.AddVV(pos , d);
						sprite_collmesh.graphics.lineTo((v.x) * PHYSICAL_SCALE, (v.y) * PHYSICAL_SCALE);
						theta += k_increment;
					}
					sprite_collmesh.graphics.lineTo((pos.x + r) * PHYSICAL_SCALE, (pos.y) * PHYSICAL_SCALE);
					
					sprite_collmesh.graphics.moveTo((pos.x) * PHYSICAL_SCALE, (pos.y) * PHYSICAL_SCALE);
					var ax:b2Vec2 = circle.m_R.col1;
					var pos2:b2Vec2 = new b2Vec2(pos.x + r * ax.x, pos.y + r * ax.y);
					sprite_collmesh.graphics.lineTo((pos2.x) * PHYSICAL_SCALE, (pos2.y) * PHYSICAL_SCALE);
				}
				break;
				
				case b2Shape.e_polyShape:
				{
					var poly:b2PolyShape = shape as b2PolyShape;
					var tV:b2Vec2 = b2Math.AddVV(poly.m_position, b2Math.b2MulMV(poly.m_R, poly.m_vertices[i]));
					sprite_collmesh.graphics.lineStyle(1,0x000000,1);
					sprite_collmesh.graphics.moveTo(tV.x * PHYSICAL_SCALE, tV.y * PHYSICAL_SCALE);
					
					for (i = 0; i < poly.m_vertexCount; ++i)
					{
						v = b2Math.AddVV(poly.m_position, b2Math.b2MulMV(poly.m_R, poly.m_vertices[i]));
						sprite_collmesh.graphics.lineTo(v.x * PHYSICAL_SCALE, v.y * PHYSICAL_SCALE);
					}
					sprite_collmesh.graphics.lineTo(tV.x * PHYSICAL_SCALE, tV.y * PHYSICAL_SCALE);
				}
				break;
			}
		}
		
		
		
		
		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		private function DrawJoint(joint:b2Joint):void
		{
			var b1:b2Body = joint.m_body1;
			var b2:b2Body = joint.m_body2;
			
			var x1:b2Vec2 = b1.m_position;
			var x2:b2Vec2 = b2.m_position;
			var p1:b2Vec2 = joint.GetAnchor1();
			var p2:b2Vec2 = joint.GetAnchor2();
			
			sprite_collmesh.graphics.lineStyle(1,0x44aaff,1/1);
			
			switch (joint.m_type)
			{
			case b2Joint.e_distanceJoint:
			case b2Joint.e_mouseJoint:
				sprite_collmesh.graphics.moveTo(p1.x * PHYSICAL_SCALE, p1.y * PHYSICAL_SCALE);
				sprite_collmesh.graphics.lineTo(p2.x * PHYSICAL_SCALE, p2.y * PHYSICAL_SCALE);
				break;
				
			case b2Joint.e_pulleyJoint:
				var pulley:b2PulleyJoint = joint as b2PulleyJoint;
				var s1:b2Vec2 = pulley.GetGroundPoint1();
				var s2:b2Vec2 = pulley.GetGroundPoint2();
				sprite_collmesh.graphics.moveTo(s1.x * PHYSICAL_SCALE, s1.y * PHYSICAL_SCALE);
				sprite_collmesh.graphics.lineTo(p1.x * PHYSICAL_SCALE, p1.y * PHYSICAL_SCALE);
				sprite_collmesh.graphics.moveTo(s2.x * PHYSICAL_SCALE, s2.y * PHYSICAL_SCALE);
				sprite_collmesh.graphics.lineTo(p2.x * PHYSICAL_SCALE, p2.y * PHYSICAL_SCALE);
				break;
				
			default:
				if (b1 == world.m_groundBody){
					sprite_collmesh.graphics.moveTo(p1.x * PHYSICAL_SCALE, p1.y * PHYSICAL_SCALE);
					sprite_collmesh.graphics.lineTo(x2.x * PHYSICAL_SCALE, x2.y * PHYSICAL_SCALE);
				}
				else if (b2 == world.m_groundBody){
					sprite_collmesh.graphics.moveTo(p1.x * PHYSICAL_SCALE, p1.y * PHYSICAL_SCALE);
					sprite_collmesh.graphics.lineTo(x1.x * PHYSICAL_SCALE, x1.y * PHYSICAL_SCALE);
				}
				else{
					sprite_collmesh.graphics.moveTo(x1.x * PHYSICAL_SCALE, x1.y * PHYSICAL_SCALE);
					sprite_collmesh.graphics.lineTo(p1.x * PHYSICAL_SCALE, p1.y * PHYSICAL_SCALE);
					sprite_collmesh.graphics.lineTo(x2.x * PHYSICAL_SCALE, x2.y * PHYSICAL_SCALE);
					sprite_collmesh.graphics.lineTo(p2.x * PHYSICAL_SCALE, p2.y * PHYSICAL_SCALE);
				}
			}
		}

	}
}