 

////////////////////////////////////////////////////////////////////////
//
// base game obj
//
////////////////////////////////////////////////////////////////////////


package si.game
{

	include "../siImportFile.as"


	public class GameObj extends Sprite
	{

		protected var vel_adj:Point;
		protected var size:Point;
		protected var pobj:Sprite;
		protected var game:Game;
		protected var delete_me:Boolean = false;
		protected var flags:uint;

		//flags		
		public static const FLAG_DISPLAYABLE:uint =    	  0x0001;			
		public static const FLAG_COLLIDABLE:uint =  	  0x0002;
		public static const FLAG_COLLIDABLE_SOLID:uint =  0x0004;
		public static const FLAG_PUSH_TO_BACK:uint =  	  0x0008;

		//collision info
		protected var collision_category:uint = 0;
		protected var collision_mask:uint = 0;
		protected var coll_body:b2Body;

		////////////////////////////////////////////////////////////////////////
		//
		// 
		//
		////////////////////////////////////////////////////////////////////////
		public function GameObj(game:Game,pobj:Sprite, pos:Point, in_size:Point, flags:uint, coll_category:uint, coll_mask:uint)
		{
		

			//set collision info
			collision_category = coll_category;
			collision_mask = coll_mask;
			
			vel_adj = new Point(0, 0);
			this.flags = flags;
			this.pobj = pobj;
			this.game = game;
			this.x = pos.x;
			this.y = pos.y;

			//size obj because cant use width and height cause they are reflective
			//of the current size of the object (i.e. if you add a child to an obj,
			//then that will affects its overall w,h
			this.size = in_size;

			this.draw();
						
			addToLists(flags);


		}
	

		///////////////////////////////////////////////////////////////////////////
		//
		//
		///////////////////////////////////////////////////////////////////////////
		public function shutdown():void
		{
			removeFromLists();
			graphics.clear();				
		}

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		private function addToLists(flags:uint):void
		{
			//display lists
			if (flags&FLAG_DISPLAYABLE)
			{
				pobj.addChild(this);
			}


			if (flags&FLAG_PUSH_TO_BACK)
			{
				pobj.setChildIndex(this,0);
			}
					

			//create collision mesh
			if (flags&FLAG_COLLIDABLE)
				coll_body = game.getCollSystem().createCollisionMesh(false, collision_category, collision_mask,size, x,y,this);
			else if (flags&FLAG_COLLIDABLE_SOLID)
				coll_body = game.getCollSystem().createCollisionMesh(true, collision_category, collision_mask,size, x,y,this);

		}


		

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		private function removeFromLists():void
		{

			//collision lists
			if (flags&FLAG_COLLIDABLE || flags&FLAG_COLLIDABLE_SOLID)
				game.getCollSystem().destroyCollisionMesh(coll_body);

			//display lists
			if (this.flags&FLAG_DISPLAYABLE)
				pobj.removeChild(this);
		
		}



		////////////////////////////////////////////////////////////////////////
		//
		// info:
		//
		//	Q: why is there a 'vel * (time_diff/1000)' ?
		//
		//	A: the game is set to run on a timer (default is 60 fps)
		//	   but it's not perfect:
		//		- somtimes one machine is slower than another
		//		- sometimes things happen in the game that cause it to 
		//		  slow it down (i.e. a super effect that uses a lot
		//		  of processing power)
		//
		//	   to solve this, objects are moved in proportion to how much
		//	   slowdown/speedup has occured.
		//
		//	   here's the math:
		//	
		//		D = R * T
		//
		//		case 1)
		//			assume a ball moves at 100 pixels per second
		//			assume that it has been 60 milliseconds since our last screen update
		//
		//			T = 60/1000 = 0.060 seconds
		//
		//			D = 100 (pixels/second) * 0.060 seconds
		//		
		//			D = 6 pixels
		//
		//			so for this frame, we should move the ball 6 pixels
		//
		//		case 2)
		//			assume a ball moves at 100 pixels per second
		//			assume that it has been 1000 milliseconds since our last screen update
		//
		//			T = 1000/1000 = 1 seconds
		//
		//			D = 100 (pixels/second) * 1 seconds
		//		
		//			D = 100 pixels
		//
		//			so for this frame, we should move the ball 100 pixels
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function update(time_diff:int):void
		{
		
		
			if (vel_adj.x != 0 || vel_adj.y != 0)
			{

				if (flags&FLAG_COLLIDABLE)
				{
					var v:b2Vec2 = coll_body.GetLinearVelocity();
					v.x += vel_adj.x;
					v.y += vel_adj.y;

					coll_body.SetLinearVelocity(v);


				}
				else
				{
					x += vel_adj.x;
					y += vel_adj.y;
				}		


				vel_adj.x = 0;
				vel_adj.y = 0;
			}
		}



		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function draw():void
		{

			graphics.clear();			

			var grfx:Graphics = this.graphics;
			grfx.moveTo(-this.size.x/2, -this.size.y/2);
			grfx.beginFill(0x000077);
			grfx.lineStyle(0, 0, 0, true);
			grfx.lineTo(this.size.x/2,-this.size.y/2);
			grfx.lineTo(this.size.x/2, this.size.y/2);
			grfx.lineTo(-this.size.x/2, this.size.y/2);
			grfx.lineTo(-this.size.x/2, -this.size.y/2);
			grfx.endFill();

		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function setPosRot(x:int, y:int, rot:Number):void
		{
			//hard set vels			
			var pos:b2Vec2 = new b2Vec2(0,0);			
			coll_body.SetLinearVelocity(pos);
			coll_body.SetAngularVelocity(0.0);

			//calc x and y based on orig position
			coll_body.m_position.x = x;
			coll_body.m_position.y = y;
			coll_body.m_rotation = rot;
			coll_body.SynchronizeShapes();
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function setRot(rot:Number):void
		{
			
			if (coll_body)
			{
				coll_body.m_rotation = rot;
				coll_body.SynchronizeShapes();
			}
			else
			{
				rotation = rot;
			}
		}

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function getRot():Number
		{			
			if (coll_body)
				return coll_body.m_rotation;
			else
				return rotation;
				
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function setVelAdj(vel:Point):void
		{
			vel_adj.x = vel.x;
			vel_adj.y = vel.y;
		}



		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function getVel():Point
		{
			var v:b2Vec2 = coll_body.GetLinearVelocity();
			
			var vel:Point = new Point(v.x,v.y);
			return vel;
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function setVel(vel:Point):void
		{
			var v:b2Vec2 = new b2Vec2(vel.x,vel.y);
			coll_body.SetLinearVelocity(v);
		}


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function setSpin(val:Number):void { coll_body.SetAngularVelocity(val); }
		public function setAngularDamping(val:Number):void { coll_body.m_angularDamping = 0.0; }


		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////			
		public function createMouseJoint():void
		{
			if (coll_body)
				game.getCollSystem().createMouseJoint(coll_body);
		}
		

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function destroyMouseJoint():void
		{
			if (coll_body)
				game.getCollSystem().destroyMouseJoint();
		}

		////////////////////////////////////////////////////////////////////////
		//
		//
		//
		////////////////////////////////////////////////////////////////////////
		public function getParentObj():Sprite { return pobj; }
		public function deleteMe():Boolean { return delete_me; }
		public function setForDelete():void { delete_me = true; }
		public function getX():int { return coll_body.m_position.x; }
		public function getY():int { return coll_body.m_position.y; }
		public function setSizeX(val:int):void { size.x = val; }
		public function setSizeY(val:int):void { size.y = val; }
		public function getCollBody():b2Body { return coll_body; }
		public function getSizeX():int { return size.x; }
		public function getSizeY():int { return size.y; }

	}
}